package frc.robot.subsystems;

import com.revrobotics.CANSparkLowLevel.MotorType;

import java.util.function.DoubleSupplier;

import com.revrobotics.CANSparkMax;
import com.revrobotics.CANSparkBase.IdleMode;

import edu.wpi.first.math.controller.ArmFeedforward;
import edu.wpi.first.math.controller.PIDController;
import edu.wpi.first.wpilibj.Encoder;
import edu.wpi.first.wpilibj2.command.Command;
import edu.wpi.first.wpilibj2.command.SubsystemBase;
import frc.robot.constants.HoodConstants;
import frc.robot.constants.HoodConstants.HoodPositions;

public class Hood extends SubsystemBase {
    private CANSparkMax shooterPivot;

    private Encoder pivotEncoder;

    private PIDController shooterPivotPID;
    private ArmFeedforward shooterPivotFF;

    private HoodPositions targetPosition;

    private boolean pidControl;

    public Hood() {
        shooterPivot = new CANSparkMax(HoodConstants.kMotorCANID, MotorType.kBrushless);
        shooterPivot.setInverted(true);

        shooterPivot.setSmartCurrentLimit(HoodConstants.kMotorCurrentLimit);

        pivotEncoder = new Encoder(0, 1);
        pivotEncoder.setDistancePerPulse(HoodConstants.kEncoderStepConversion);

        shooterPivot.burnFlash();

        shooterPivotPID = new PIDController(
            HoodConstants.kP,
            HoodConstants.kI,
            HoodConstants.kD
        );

        shooterPivotFF = new ArmFeedforward(
            HoodConstants.kSFF,
            HoodConstants.kGFF,
            HoodConstants.kVFF
        );

        shooterPivot.setIdleMode(IdleMode.kBrake);

        targetPosition = HoodPositions.kLoadAngle;

        pidControl = true;
    }

    @Override
    public void periodic() {
        if(pidControl) {
            shooterPivot.setVoltage(
                shooterPivotPID.calculate(
                    getShooterAngle(),
                    targetPosition.getAngle()
                ) + shooterPivotFF.calculate(
                    targetPosition.getAngle(), 
                    0.0
                )
            );
        }
    }

    public Command setHoodAngle(HoodPositions newPosition) {
        return runOnce(() -> {
            pidControl = true;

            targetPosition = newPosition;
        });
    }

    public Command setSpeed(DoubleSupplier speed) {
        return runOnce(() -> { pidControl = false; }).andThen(run(() -> {
            shooterPivot.set(speed.getAsDouble());
        }));
    }

    public Command climbState() {
        return runOnce(() -> {
            pidControl = false;
            shooterPivot.setIdleMode(IdleMode.kCoast);
        }).andThen(run(() -> {
            shooterPivot.set(0);
        })).finallyDo(() -> {
            // finallyDo will reset break mode if we back out of climbing and interrupt
            // the climbState command with something else
            shooterPivot.setIdleMode(IdleMode.kBrake);
        });
    }

    public Command stop() {
        return setSpeed(() -> 0);
    }

    public double getShooterAngle() {
        return pivotEncoder.getDistance() + HoodPositions.kLoadAngle.getAngle();
    }
}
